package CSU.module.algorithm;

import adf.agent.communication.MessageManager;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.module.algorithm.PathPlanning;
//import csu.model.route.pov.CostFunction;
//import csu.model.route.pov.CostFunctionFactory;
//import csu.model.route.pov.POVSearch;
//import csu.model.route.pov.graph.PassableDictionary;
//import csu.model.route.pov.graph.PointOfVisivility;
//import csu.model.route.pov.graph.RootNode;
//import csu.model.route.pov.reachable.UFTReachableArea;
import rescuecore2.misc.collections.LazyMap;
import rescuecore2.standard.entities.Area;
//import rescuecore2.standard.entities.Human;
//import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

//import CSU_YunLu.module.helper.WorldComputer;

public class AStarPathPlanning extends PathPlanning {

	private Map<EntityID, Set<EntityID>> graph;

	private EntityID from;
	private Collection<EntityID> targets;
	private List<EntityID> result;

	// private PassableDictionary passableDic;
	// private BFSReachableArea bfsReachable;
	// private UFTReachableArea uftReachable;
	// private PointOfVisivility pov;
	// private POVSearch search;

	// private CostFunction normalFunc;
	// private CostFunction strictFunc;
	// private CostFunction searchFunc;
	// private CostFunction pfFunc;
	// private CostFunction cost;
	//
	// private double routeCost;

	public AStarPathPlanning(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager,
			DevelopData developData) {
		super(ai, wi, si, moduleManager, developData);
		this.init();
	}

	private void init() {
		Map<EntityID, Set<EntityID>> neighbours = new LazyMap<EntityID, Set<EntityID>>() {
			@Override
			public Set<EntityID> createValue() {
				return new HashSet<>();
			}
		};
		for (Entity next : this.worldInfo) {
			if (next instanceof Area) {
				Collection<EntityID> areaNeighbours = ((Area) next).getNeighbours();
				neighbours.get(next.getID()).addAll(areaNeighbours);
			}
		}
		this.graph = neighbours;
		// pov = new PointOfVisivility(worldInfo);
		// passableDic = new PassableDictionary(agentInfo, worldInfo,
		// scenarioInfo);
		// bfsReachable = new BFSReachableArea(world);
		// uftReachable = new UFTReachableArea(worldInfo);

		// final int routeThinkTime = (int)
		// scenarioInfo.getKernelAgentsThinkTime() * 2 / 3;
		// search = new POVSearch(routeThinkTime, pov);

		// normalFunc = CostFunctionFactory.normal(worldInfo, passableDic);
		// strictFunc = CostFunctionFactory.strict(worldInfo, passableDic);
		// searchFunc = CostFunctionFactory.search(worldInfo, passableDic);
		// cost = CostFunctionFactory;

		// pfFunc = CostFunctionFactory.pf(worldInfo);
	}

	@Override
	public List<EntityID> getResult() {
		return this.result;
	}

	@Override
	public PathPlanning setFrom(EntityID id) {
		this.from = id;
		return this;
	}

	@Override
	public PathPlanning setDestination(Collection<EntityID> targets) {
		this.targets = targets;
		return this;
	}

	@Override
	public PathPlanning updateInfo(MessageManager messageManager) {
		super.updateInfo(messageManager);
		return this;
	}

	@Override
	public PathPlanning precompute(PrecomputeData precomputeData) {
		super.precompute(precomputeData);
		return this;
	}

	@Override
	public PathPlanning resume(PrecomputeData precomputeData) {
		super.resume(precomputeData);
		return this;
	}

	@Override
	public PathPlanning preparate() {
		super.preparate();
		return this;
	}

	// public PathPlanning setCostFuc(CostFunction cost) {
	// // this.cost = cost;
	// return this;
	// }

	@Override
	public PathPlanning calc() {
		List<EntityID> open = new LinkedList<>();
		// Set<StandardEntity> tars = new HashSet<>();
		// for (EntityID id : targets)
		// tars.add(worldInfo.getEntity(id));
		// open = search.getAStarPath(worldInfo, pov.get(from), tars,
		// cost).getRoute();
		Map<EntityID, EntityID> ancestors = new HashMap<>();
		open.add(this.from);
		EntityID next;
		boolean found = false;
		ancestors.put(this.from, this.from);
		do {
			next = open.remove(0);
			if (isGoal(next, targets)) {
				found = true;
				break;
			}
			Collection<EntityID> neighbours = graph.get(next);
			if (neighbours.isEmpty()) {
				continue;
			}
			for (EntityID neighbour : neighbours) {
				if (isGoal(neighbour, targets)) {
					ancestors.put(neighbour, next);
					next = neighbour;
					found = true;
					break;
				} else {
					if (!ancestors.containsKey(neighbour)) {
						open.add(neighbour);
						ancestors.put(neighbour, next);
					}
				}
			}
		} while (!found && !open.isEmpty());
		if (!found) {
			// No path
			this.result = null;
		}
		// Walk back from goal to this.from
		EntityID current = next;
		LinkedList<EntityID> path = new LinkedList<>();
		do {
			path.add(0, current);
			current = ancestors.get(current);
			if (current == null) {
				throw new RuntimeException("Found a node with no ancestor! Something is broken.");
			}
		} while (current != this.from);
		this.result = path;
		return this;
	}

	private boolean isGoal(EntityID e, Collection<EntityID> test) {
		return test.contains(e);
	}
}
